from .client import APIClient, ProductionAPIClient, DevelopmentAPIClient, RtemClient  # noqa: F401
from .exceptions import OnboardApiException, OnboardTemporaryException

OnboardClient = ProductionAPIClient

__all__ = [
    'OnboardClient',
    'APIClient',
    'RtemClient',
    'OnboardApiException',
    'OnboardTemporaryException'
]



from typing import Iterable, Dict, List, Union

import pandas as pd

from onboard.client.models import PointData


def points_df_from_timeseries(timeseries, points=[]) -> pd.DataFrame:
    """Returns a pandas dataframe from the results of a timeseries query"""

    # 'type' is from the point_type.display_name column
    point_names = {str(p['id']): p.get('type') for p in points}

    columns = ['timestamp']
    dates = set()
    data_by_point = {}

    for point in timeseries:
        point_id = point['tags']['point_id']
        columns.append(point_id)
        col_indexes = point['columns']

        ts_index = col_indexes.index('time')
        clean_index = col_indexes.index('clean')

        point_data: Dict[str, float] = {}
        data_by_point[point_id] = point_data

        for val in point['values']:
            ts = val[ts_index]
            dates.add(ts)
            clean = val[clean_index]
            point_data[ts] = clean

    sorted_dates = list(dates)
    sorted_dates.sort()
    data = []

    for d in sorted_dates:
        row = {'timestamp': d}
        for p in columns[1:]:
            val = data_by_point[p].get(d)
            point_name = point_names.get(p)
            point_col = f"{point_name} - {p}" if point_name else p
            row[point_col] = val
        data.append(row)

    df = pd.DataFrame(data)
    return df


def points_df_from_streaming_timeseries(timeseries: Iterable[PointData],
                                        points=[],
                                        point_column_label=None,
                                        ) -> pd.DataFrame:
    """Returns a pandas dataframe from the results of a timeseries query"""
    if point_column_label is None:
        def point_column_label(p):
            return p.get('id')

    point_names = {p['id']: point_column_label(p) for p in points}
    columns: List[Union[str, int]] = ['timestamp']
    dates = set()
    data_by_point = {}

    for point in timeseries:
        columns.append(point.point_id)
        ts_index = point.columns.index('time')
        data_index = point.columns.index(point.unit)

        point_data: Dict[str, Union[str, float, None]] = {}
        data_by_point[point.point_id] = point_data

        for val in point.values:
            ts: str = val[ts_index]  # type: ignore[assignment]
            dates.add(ts)
            clean = val[data_index]
            point_data[ts] = clean

    sorted_dates = list(dates)
    sorted_dates.sort()
    data = []

    for d in sorted_dates:
        row = {'timestamp': d}
        for p in columns[1:]:
            val = data_by_point[p].get(d)  # type: ignore
            point_col = point_names.get(p, p)
            row[point_col] = val  # type: ignore
        data.append(row)

    df = pd.DataFrame(data)
    return df


def df_time_index(df: pd.DataFrame,
                  time_col='timestamp', utc=True) -> pd.DataFrame:
    dt_series = pd.to_datetime(df[time_col], infer_datetime_format=True)
    datetime_index = pd.DatetimeIndex(dt_series.values)
    if utc:
        datetime_index = datetime_index.tz_localize('UTC')
    df_indexed = df.set_index(datetime_index)
    df_indexed.drop(time_col, axis=1, inplace=True)
    return df_indexed


def df_objs_to_numeric(df: pd.DataFrame):
    cols = df.columns[df.dtypes.eq('object')]
    return df[cols].apply(pd.to_numeric, errors='coerce')



from typing import List, Iterable, TypeVar, Callable

import requests

from .exceptions import OnboardApiException, OnboardTemporaryException

T = TypeVar('T')


def divide_chunks(input_list: List[T], n: int) -> Iterable[List[T]]:
    # looping till length input_list
    for i in range(0, len(input_list), n):
        yield input_list[i:i + n]


def json(func: Callable[..., T]) -> Callable[..., T]:
    """Decorator for making sure requests responses are handled consistently"""

    # the type annotations on json are a lie to let us type the methods in client
    # with approximate descriptions of the JSON they return, even though the methods
    # as implemented return requests.Response objects
    def wrapper(*args, **kwargs):
        try:
            res: requests.Response = func(*args, **kwargs)  # type: ignore[assignment]
            if res is None:
                return None

            # remove the cached access token if authorization failed
            # it's likely just expired
            if res.status_code == 401 and args and args[0].token is not None:
                args[0].token = None
                return wrapper(*args, **kwargs)

            if res.status_code > 499:
                raise OnboardTemporaryException(res.text or res.status_code)
            if res.status_code > 399:
                raise OnboardApiException(res.text or res.status_code)

            if hasattr(wrapper, 'raw_response'):
                return res
            return res.json()
        except OnboardApiException as e:
            raise e
        except Exception as e:
            raise OnboardApiException(e)

    return wrapper



from typing import List, Dict, Any, Optional

from urllib3.util.retry import Retry

from .helpers import ClientBase
from .util import json


class StagingClient(ClientBase):
    """Staging area API bindings"""

    def __init__(self, api_url: str,
                 api_key: Optional[str] = None,
                 token: Optional[str] = None,
                 name: str = '',
                 retry: Optional[Retry] = None,
                 ) -> None:
        super().__init__(api_url, user=None, pw=None, api_key=api_key, token=token, name=name,
                         retry=retry)

    @json
    def get_staging_building_details(self) -> List[Dict]:
        """Fetch building-level details for all buildings in staging"""
        return self.get('/staging')

    @json
    def update_building_details(self, building_id: int,
                                details: Dict[str, Any]) -> Dict:
        """Update building-level details for a building in staging"""
        return self.patch(f"/staging/{building_id}/details", json=details)

    @json
    def get_staged_equipment(self, building_id: int) -> Dict:
        """Fetch staging equipment (point topic strings only) as Python objects"""
        return self.get(f'/staging/{building_id}')

    @json
    def get_equipment_and_points(self, building_id: int) -> Dict:
        """Fetch staging equipment and point details together as Python objects"""
        return self.get(f'/staging/{building_id}?points=true')

    def get_staged_equipment_csv(self, building_id: int) -> str:
        """Fetch staged equipment and points together in tabular form"""

        @json
        def get_csv():
            return self.get(f'/staging/{building_id}',
                            headers={'Accept': 'text/csv'})

        get_csv.raw_response = True  # type: ignore[attr-defined]
        return get_csv().text

    @json
    def update_staged_equipment(self, building_id: int, updates: List[Dict]) -> Dict:
        """Update staged equipment and points"""
        return self.post(f'/staging/{building_id}', json=updates)

    @json
    def validate_staging_building(self, building_id: int) -> Dict:
        """Validate staged equipment and points, returning any errors"""
        return self.get(f'/staging/{building_id}/validate')

    @json
    def promote_from_staging(self, building_id: int,
                             equip_ids: List[str] = [], topics: List[str] = []) -> Dict:
        """Promote valid equipment and points to the primary tables, returning any errors
        If equip_ids or topics lists are non-empty then only promote those objects. Otherwise
        all valid objects are promoted."""
        promote_req = {'equip_ids': equip_ids, 'topics': topics}
        return self.post(f'/staging/{building_id}/apply', json=promote_req)


class OnboardStagingClient(StagingClient):
    def __init__(self, api_key: str) -> None:
        super().__init__('https://api.onboarddata.io', api_key)



from setuptools import setup

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open("README.md", "r") as f:
    long_description = f.read()

setup(name='onboard.client',
      version='1.10.2',
      author='Nathan Merritt, John Vines',
      author_email='support@onboarddata.io',
      description='Onboard API SDK',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/onboard-data/client-py',
      packages=['onboard.client'],
      install_requires=requirements,
      package_data={
          'onboard.client': ['py.typed'],
      },
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11',
          'Topic :: Scientific/Engineering :: Information Analysis',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      python_requires='>=3.8',
      )



import urllib.parse
from datetime import datetime
from typing import List, Dict, Any, Optional, Tuple, Union, Iterator

import deprecation
from orjson import loads
from urllib3.util.retry import Retry

from .exceptions import OnboardApiException
from .helpers import ClientBase
from .models import PointSelector, PointDataUpdate, IngestStats, \
    TimeseriesQuery, PointData
from .util import divide_chunks, json


class APIClient(ClientBase):
    def __init__(self,
                 api_url: str,
                 user: Optional[str] = None,
                 pw: Optional[str] = None,
                 api_key: Optional[str] = None,
                 token: Optional[str] = None,
                 name: str = '',
                 retry: Optional[Retry] = None,
                 ) -> None:
        super().__init__(api_url, user, pw, api_key, token, name, retry)

    @json
    def whoami(self) -> Dict[str, str]:
        """returns the current account's information"""
        return self.get('/whoami')

    @json
    def get_account_actions(self) -> List[Dict[str, str]]:
        """returns the action audit log by or affecting the current account"""
        return self.get('/account-actions')

    @json
    def get_users(self) -> List[Dict[str, str]]:
        """returns the list of visible user accounts
          For organization admins this is all users in the organization
          For non-admin users this is just the current account
        """
        return self.get('/users')

    @json
    def get_organizations(self) -> Dict[str, Dict[str, str]]:
        return self.get('/organizations')

    @json
    def get_all_buildings(self) -> List[Dict[str, str]]:
        return self.get('/buildings')

    @json
    def get_tags(self) -> List[Dict[str, str]]:
        """returns a list of all the haystack tags in the system
        For more info, please see https://project-haystack.org/tag"""
        return self.get('/tags')

    @json
    def get_equipment_types(self) -> List[Dict[str, str]]:
        return self.get('/equiptype')

    @json
    def get_building_equipment(self, building_id: int) -> List[Dict[str, Any]]:
        return self.get(f'/buildings/{building_id}/equipment?points=true')

    @json
    def get_equipment_by_ids(self, equipment_ids: List[int]) -> List[Dict[str, object]]:
        body = {'equipment_ids': equipment_ids}
        return self.post('/equipment/query', json=body)

    @json
    def get_building_changelog(self, building_id: int) -> List[Dict[str, object]]:
        """Returns a list of changelog entries for the specified building"""
        return self.get(f'/buildings/{building_id}/changelog')

    @json
    def select_points(self, selector: PointSelector) -> Dict[str, List[int]]:
        """returns point ids based on the provided selector"""
        return self.post('/points/select', json=selector.json())

    def check_data_availability(self,
                                selector: PointSelector
                                ) -> Tuple[Optional[datetime], Optional[datetime]]:
        """Returns a tuple of data timestamps (most stale, most recent) for selected points"""

        @json
        def get_as_json():
            return self.post('/points/data-availability', json=selector.json())

        res = get_as_json()
        oldest = self.ts_to_dt(res['oldest'])
        newest = self.ts_to_dt(res['newest'])
        return (oldest, newest)

    def get_all_points(self) -> List[Dict[str, Any]]:
        """returns all points for all visible buildings"""
        buildings = self.get_all_buildings()
        points: List[Dict[str, Any]] = []
        for b in buildings:
            bldg_id = b['id']
            equipment = self.get_building_equipment(bldg_id)
            for e in equipment:
                points += e['points']
        return points

    def get_all_equipment(self) -> List[Dict]:
        """returns all equipment instances for all visible buildings"""
        buildings = self.get_all_buildings()
        equipment = []
        for b in buildings:
            bldg_id = b['id']
            equipment += self.get_building_equipment(bldg_id)
        return equipment

    def get_points_by_ids(self, point_ids: List[int]) -> List[Dict[str, str]]:
        @json
        def get_points(url):
            return self.get(url)

        points = []
        for chunk in divide_chunks(point_ids, 500):
            points_str = '[' + ','.join(str(id) for id in chunk) + ']'
            url = f'/points?point_ids={points_str}'
            try:
                points_chunk = get_points(url)
            except OnboardApiException as e:
                if '"status": 404' in str(e):
                    continue
                raise e
            points += points_chunk
        return points

    # Deprecated
    def get_points_by_datasource(self, datasource_hashes: List[str]) -> List[Dict[str, str]]:
        datasource_hashes_chunked = list(divide_chunks(datasource_hashes, 125))

        @json
        def get_points(url):
            return self.get(url)

        points = []
        for chunk in datasource_hashes_chunked:
            hashes_str = "[" + ','.join([r"'" + c + r"'" for c in chunk]) + "]"
            query = urllib.parse.quote(hashes_str)
            url = f'/points?datasource_hashes={query}'
            try:
                points_chunk = get_points(url)
            except OnboardApiException as e:
                if '"status": 404' in str(e):
                    continue
                raise e
            points += points_chunk
        return points

    @json
    def get_all_point_types(self) -> List[Dict[str, str]]:
        return self.get('/pointtypes')

    @json
    def get_all_measurements(self) -> List[Dict[str, str]]:
        return self.get('/measurements')

    @json
    def get_all_units(self) -> List[Dict[str, str]]:
        return self.get('/unit')

    @deprecation.deprecated(deprecated_in="1.3.0",
                            details="Use stream_point_timeseries instead")
    @json
    def query_point_timeseries(self, point_ids: List[int],
                               start_time: Union[str, datetime],
                               end_time: Union[str, datetime]) -> List[Dict[str, Any]]:
        """Query a timespan for a set of point ids
        point_ids: a list of point ids
        start/end time: ISO formatted timestamp strings e.g. '2019-11-29T20:16:25Z' or a datetime
        """
        query = {
            'point_ids': point_ids,
            'start_time': self.dt_to_str(start_time),
            'end_time': self.dt_to_str(end_time),
        }
        return self.post('/query', json=query)

    def stream_point_timeseries(self, query: TimeseriesQuery) -> Iterator[PointData]:
        """Query a time interval for an explicit set of point ids or
        with a selector which describes which sensors to include.

        Example values docmentaed on the model tab here:
            https://api.onboarddata.io/doc/#/buildings%3Aread/post_query_v2
        """

        @json
        def query_call():
            return self.post('/query-v2', json=query.json(), stream=True,
                             headers={'Accept': 'application/x-ndjson'})

        query_call.raw_response = True  # type: ignore[attr-defined]

        try:
            # Pydantic v1
            point_data = PointData.__pydantic_model__.construct  # type: ignore[attr-defined]
        except AttributeError:
            # Pydantic v2
            point_data = PointData.model_construct  # type: ignore[attr-defined]

        with query_call() as res:
            for line in res.iter_lines(chunk_size=20 * 1024):
                parsed = loads(line)
                yield point_data(**parsed)

    @json
    def update_point_data(self, updates: List[PointDataUpdate] = []) -> None:
        """Bulk update point data, returns the number of updated points
        updates: an iterable of models.PointDataUpdate objects"""
        for batch in divide_chunks(updates, 500):
            json = [u.json() for u in batch]
            self.post('/points_update', json=json).raise_for_status()

    @json
    def send_ingest_stats(self, ingest_stats: IngestStats) -> None:
        """Send timing and diagnostic info to the portal
        ingest_stats: an instance of models.IngestStats"""
        json = ingest_stats.json()
        self.post('/ingest-stats', json=json).raise_for_status()

    @json
    def get_ingest_stats(self) -> List[Dict[str, str]]:
        """returns ingest stats for all buildings"""
        return self.get('/ingest-stats')

    @json
    def get_alerts(self) -> List[Dict[str, str]]:
        """returns a list of active alerts for all buildings"""
        return self.get('/alerts')

    @json
    def copy_point_data(self, point_id_map: Dict[int, int],
                        start_time: Union[str, datetime],
                        end_time: Union[str, datetime]) -> str:
        """Copy data between points
        point_id_map: a map of source to destination point id
        start/end: ISO formatted timestamp strings e.g. '2019-11-29T20:16:25Z'
        returns: a string describing the operation
        """
        command = {
            'point_id_map': point_id_map,
            'start_time': self.dt_to_str(start_time),
            'end_time': self.dt_to_str(end_time),
        }
        return self.post('/point-data-copy', json=command)


class DevelopmentAPIClient(APIClient):
    def __init__(self,
                 user: Optional[str] = None,
                 pw: Optional[str] = None,
                 api_key: Optional[str] = None,
                 token: Optional[str] = None,
                 retry: Optional[Retry] = None,
                 ) -> None:
        super().__init__('https://devapi.onboarddata.io', user, pw, api_key, token, retry=retry)


class ProductionAPIClient(APIClient):
    def __init__(self,
                 user: Optional[str] = None,
                 pw: Optional[str] = None,
                 api_key: Optional[str] = None,
                 token: Optional[str] = None,
                 retry: Optional[Retry] = None,
                 ) -> None:
        super().__init__('https://api.onboarddata.io', user, pw, api_key, token, retry=retry)


class RtemClient(APIClient):
    def __init__(self,
                 api_key: Optional[str] = None,
                 retry: Optional[Retry] = None,
                 ) -> None:
        super().__init__('https://api.ny-rtem.com', api_key=api_key, retry=retry)
